from .constants import VALID_ENCRYPTION_METHODS
class InvalidSettingsProvided(Exception):
    pass


class NoDefaultMapTagDefinedException(Exception):
    def __init__(self, msg="No default section is defined"):
        super().__init__(msg)


class TooManyDefaultMapTagsDefinedException(Exception):
    def __init__(self, msg="Too many default sections defined"):
        super().__init__(msg)


class EnvironmentNotFound(Exception):
    def __init__(self, msg="Environment not found"):
        super().__init__(msg)


class NoEnvironmentsDefinedException(Exception):
    def __init__(self, msg="No Environments defined, at least 1 !env must be defined"):
        super().__init__(msg)


class EnvironmentHasNoEncryptedSecretTagsException(Exception):
    def __init__(
        self, msg="Environment does not have any !encrypted tags, cannot decrypt"
    ):
        super().__init__(msg)


class EnvironmentHasNoSecretTagsException(Exception):
    def __init__(self, msg="Environment does not have any !secrets, cannot encrypt"):
        super().__init__(msg)


class EnvironmentIsAlreadyEncrypted(Exception):
    def __init__(self, msg="Environment is already encrypted"):
        super().__init__(msg)


class EnvironmentIsAlreadyDecrypted(Exception):
    def __init__(self, msg="Environment is already decrypted"):
        super().__init__(msg)


class VersionTagNotSpecified(Exception):
    def __init__(
        self, msg="No version tag has been specified, please add version: 1.0"
    ):
        super().__init__(msg)


class UnsupportedVersionSpecified(Exception):
    def __init__(self, msg="Only version 1.0 is currently supported"):
        super().__init__(msg)


class UnsupportedEncryptionMethodSpecified(Exception):
    def __init__(self, encryption_method):
        encryption_methods = ', '.join(VALID_ENCRYPTION_METHODS)
        msg = f"Invalid encryption method specified: {encryption_method}, must be one of {encryption_methods}"
        super().__init__(msg)



import os
from typing import List
from yamlvault.processor import SecretYAML


def load_settings_from_config(
    config_file_path: str,
    environment: str,
    passwords: List = None,
    password_files: List = None,
):
    if not os.path.isfile(config_file_path):
        raise FileNotFoundError(f"{config_file_path} found not be found")

    config = SecretYAML(filepath=config_file_path)
    # TODO: check if yaml has the specified environment

    if config.is_default_decrypted() and config.is_env_decrypted(environment):
        # returns the config as a dict, when neither default or environment is encrypted
        return config.get_env_as_dict(environment)

    if config.is_default_encrypted():
        if not passwords and not password_files:
            raise Exception("No passwords or password files specified")

        if passwords:
            for password in passwords:
                config.decrypt_default(password, raise_exception=False)

        if password_files:
            for password_file in password_files:
                if not os.path.isfile(password_file):
                    raise FileNotFoundError(f"{password_file} found not be found")
                password = open(password_file, "r").read()
                config.decrypt_default(password, raise_exception=False)

    if config.is_env_encrypted(environment):
        if not passwords and not password_files:
            raise Exception("No passwords or password files specified")

        if passwords:
            for password in passwords:
                config.decrypt_env(environment, password, raise_exception=False)

        if password_files:
            for password_file in password_files:
                if not os.path.isfile(password_file):
                    raise FileNotFoundError(f"{password_file} found not be found")
                password = open(password_file, "r").read()
                config.decrypt_env(environment, password, raise_exception=False)

    return config.get_env_as_dict(environment)



from environ import Env
from django.conf import settings


def generate_dynamic_environ(config):
    env = Env()
    env.read_env()
    settings.configure({})
    # Set environment variables dynamically
    for key, value in list(config.items()):
        # Determine the type of the value
        if isinstance(value, bool):
            # For booleans
            setattr(settings, key, env.bool(key, default=value))
        elif isinstance(value, int):
            setattr(settings, key, env.int(key, default=value))
        elif isinstance(value, float):
            # For floats
            setattr(settings, key, env.float(key, default=value))
        else:
            # Default to string
            var = env.str(key, default=str(value))
            setattr(settings, key, var)



CONTAINS_ENCRYPTED_TAGS = "yaml contains encrypted tags"
CONTAINS_UNENCRYPTED_TAGS = "yaml contains unencrypted tags"


ANSIBLE_VAULT_ENCRYPTION_METHOD = 'ansible-vault'

VALID_ENCRYPTION_METHODS = [
    ANSIBLE_VAULT_ENCRYPTION_METHOD
]


from ruamel.yaml.nodes import ScalarNode


class OtherScalar:
    style = None
    tag = None
    value = None
    parent = None
    parent_name = None

    def __init__(self, tag, value, style=None):
        self.tag = tag
        self.value = value
        self.style = style

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str", data.value, style=data.style
        )  # Represent as lowercase in YAML

    @classmethod
    def from_yaml(cls, constructor, node):
        if node.style is None:
            return node.value
        return node.value, cls(node.tag, node.value, style=node.style)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(self.value)


class DefaultSecretConfigMap:
    yaml_tag = "!default"
    parent = None
    parent_name = None

    def __init__(self, tag, value):
        self.tag = tag
        self.value = value

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag, node.value)

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(node.tag, node.value)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()


class EnvSecretConfigMap:
    yaml_tag = "!env"
    style = None
    parent = None
    parent_name = None

    def __init__(self, tag, value, style=None):
        self.tag = tag
        self.value = value
        self.style = style

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag, node.value, style=node.style)

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(node.tag, node.value, style=node.style)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(f"{self.yaml_tag} {self.value}")


class SecretString:
    yaml_tag = "!secret"
    alias_key = None
    style = None
    parent = None
    parent_name = None

    def __init__(self, value, style=None):
        self.value = value
        self.style = style

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag, node.value, style=node.style)

    @classmethod
    def from_yaml(cls, constructor, node):
        return cls(node.value, style=node.style)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(f"{self.yaml_tag} {self.value}")


class EncryptedString:
    yaml_tag = "!encrypted"
    style = None
    parent = None
    parent_name = None

    def __init__(self, value, style=None):
        self.value = value
        self.style = style

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag, node.value, style=node.style)

    @classmethod
    def from_yaml(cls, constructor, node):
        loader = constructor.loader
        return cls(node.value, style=node.style)

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return str(f"{self.yaml_tag} {self.value}")


class RequiredString:
    yaml_tag = "!required"
    style = None
    parent = None
    parent_name = None

    def __init__(self, value, style=None):
        self.value = value
        self.style = style

    @classmethod
    def to_yaml(cls, representer, node):
        return representer.represent_scalar(cls.yaml_tag, node.value, style=node.style)

    @classmethod
    def from_yaml(cls, constructor, node):
        loader = constructor.loader
        return cls(node.value)


class StyledScalar:
    parent = None
    parent_name = None

    def __init__(self, value, style):
        self.value = value
        self.style = style

    def __hash__(self):
        return hash((self.value, self.style))

    def __eq__(self, other):
        if isinstance(other, StyledScalar):
            return self.value == other.value and self.style == other.style
        return False

    def __repr__(self):
        return str(self.value)

    def __str__(self):
        return str(self.value)


def scalar_constructor(loader, node):
    assert isinstance(node, ScalarNode)
    style = node.style  # None (plain), ' (single-quoted), or " (double-quoted)
    value = loader.construct_scalar(node)
    return StyledScalar(value, style)


def scalar_representer(dumper, data):
    if isinstance(data, StyledScalar):
        return dumper.represent_scalar(
            "tag:yaml.org,2002:str", data.value, style=data.style
        )



import os
import sys
import logging
import pprint

import click

from yamlvault.processor import SecretYAML

logger = logging.getLogger(__name__)
logger.level = logging.INFO

@click.group()
def main():
    pass

@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.argument('config')
@click.argument('env', default="default")
@click.option('-p', '--password',  help='password for specified env, you can specify it multiple times')
@click.option('-pf', '--password-file',  help='password file for specified env, you can specify it multiple times')
@click.option('--dryrun', is_flag=True, default=False, help='Dry run verbose mode')
@click.option('-v', '--verbose', is_flag=True, default=False, help='Enables verbose mode')
def encrypt(config, env, password, password_file, dryrun, verbose):
    if password_file:
        if not os.path.isfile(pf):
            raise FileNotFoundError(pf)
        with open(pf, "r") as file:
            password = file.read().strip()

    if not os.path.isfile(config):
        raise FileNotFoundError(configg)

    config = SecretYAML(filepath=config)
    if env == "default":
        config.encrypt_default(password)
    else:
        config.encrypt_env(env, password)

    if not dryrun:
        config.save_file()
        click.echo(f"Encrypted {env} in {config}")
        output = pprint.pformat(config.to_dict(), indent=4, width=1)
        if verbose:
            click.echo(output)
    else:
        output = pprint.pformat(config.to_dict(), indent=4, width=1)
        click.echo(f"[dry-run] Encrypted {env} in {config}")
        if verbose:
            click.echo(output)


@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.argument('config')
@click.argument('env', default="default")
@click.option('-p', '--password',  help='password for specified env')
@click.option('-pf', '--password-file',  help='password file for specified env')
@click.option('--dryrun', is_flag=True, default=False, help='Dry run verbose mode')
@click.option('-v', '--verbose', is_flag=True, default=False, help='Enables verbose mode')
def decrypt(config, env, password, password_file, dryrun, verbose):
    if password_file:
        if not os.path.isfile(password_file):
            raise FileNotFoundError(password_filee)
        with open(password_file, "r") as file:
            password = file.read().strip()

    if not os.path.isfile(config):
        raise FileNotFoundError(config)

    config = SecretYAML(filepath=config)
    if env == "default":
        config.decrypt_default(password)
    else:
        config.decrypt_env(env, password)

    if not dryrun:
        config.save_file()
        output = pprint.pformat(config.to_dict(), indent=4, width=1)
        click.echo(f"Decrypted {env} in {config}")
        if verbose:
            click.echo(output)
    else:
        output = pprint.pformat(config.to_dict(), indent=4, width=1)
        if verbose:
            click.echo(output)
        click.echo(f"[dry-run] Decrypted {env} in {config}")


@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.argument('config')
def dump():
    password = None
    if args.password:
        password = args.password
    elif args.password_file:
        if not os.path.isfile(args.password_file):
            raise FileNotFoundError(args.password_file)
        with open(args.password_file, "r") as file:
            password = file.read().strip()

    # env
    env = "default"
    if args.env:
        env = args.env

    if not os.path.isfile(args.config):
        raise FileNotFoundError(args.config)

    config = SecretYAML(filepath=args.config)
    pp = pprint.PrettyPrinter(depth=5, stream=None)

    if env == "default":
        config.decrypt_default(password)
        pp.pprint(config.get_default_as_dict())
    else:
        config.decrypt_env(env, password)
        pp.pprint(config.get_env_as_dict(env))

@click.command(context_settings=dict(
    ignore_unknown_options=True,
))
@click.argument('config')
def verify():
    pass


main.add_command(encrypt)
main.add_command(decrypt)
main.add_command(verify)
main.add_command(dump)


from typing import (
    Any,
    Dict,
    TypeVar,
)

from ansible.parsing.vault import VaultLib, VaultSecret, AnsibleVaultError
from ansible.constants import DEFAULT_VAULT_ID_MATCH

KeyType = TypeVar("KeyType")


def deep_update(
    mapping: Dict[KeyType, Any], *updating_mappings: Dict[KeyType, Any]
) -> Dict[KeyType, Any]:
    updated_mapping = mapping.copy()
    for updating_mapping in updating_mappings:
        for k, v in updating_mapping.items():
            if (
                k in updated_mapping
                and isinstance(updated_mapping[k], dict)
                and isinstance(v, dict)
            ):
                updated_mapping[k] = deep_update(updated_mapping[k], v)
            else:
                updated_mapping[k] = v
    return updated_mapping


def encrypt_value(value, password, node):
    try:
        vault = VaultLib([(DEFAULT_VAULT_ID_MATCH, VaultSecret(password.encode()))])
        encrypted_value = vault.encrypt(value.encode())
        return encrypted_value.decode().replace(
            "\n", "|"
        )  # replace new lines with pipe
    except Exception as e:
        print(f"Error during encryption: {e}")
        raise AnsibleVaultError(f"Could not encrypt node: {node}")


def decrypt_value(value, password, node, raise_exception=True):
    try:
        vault = VaultLib([(DEFAULT_VAULT_ID_MATCH, VaultSecret(password.encode()))])
        decrypted_value = vault.decrypt(
            value.replace("|", "\n").encode()
        )  # replace pipe with new lines
        return decrypted_value.decode()
    except Exception as e:
        print(f"Error during decryption: {e}")
        if raise_exception:
            raise AnsibleVaultError(f"Could not encrypt node: {node}")


def flatten_list(matrix):
    flat_list = []
    for row in matrix:
        flat_list.extend(row)
    return flat_list



# from typing import List, Dict
#
#
# class SecretYAMLSettings:
#     def __init__(self, settings, config_path):
#         self.settings = self.check_settings(settings)
#         self.config = self.check_config(config_path)
#
#     def check_settings(self, settings):
#         if str(type(settings)) is not 'django.conf.LazySettings':
#             raise InvalidSettingsProvided('Must be of django.conf.LazySettings type')
#         return settings
#
#     def check_config(self, config_path):
#         if not config_path.endswith('.yml'):
#             raise ValueError(f"Invalid config_path {config_path} must end with .yml")
#         try:
#             open(config['config_path'], 'r')
#         except FileNotFoundError:
#             raise ValueError(f"The file {config['config_path']} does not exist")
#         return True
#
#     def patch_settings(self, config_name, password, base_name=None):
#         # find the config from self.configs with the name config_name
#         # else return CouldNotFindConfig of config_name
#



#################################################
#   [2024] Dan Brosnan dpjbrosnan@gmail.com     #
#################################################

from __future__ import absolute_import, division, print_function

__metaclass__ = type

DOCUMENTATION = """
    vars: encrypted_config_vars
    version_added: "2.10"
    short_description: Load custom host vars from encrypted config file
    description: Load custom host vars from encrypted config file
    options:
      config_file_path:
        ini:
          - key: config_file_path
            section: encrypted_django_settings
      stage:
        ini:
          - key: stage
            section: encrypted_django_settings
    extends_documentation_fragment:
      - vars_plugin_staging
"""

import os
import argparse

from ansible.errors import AnsibleParserError
from ansible.module_utils.common.text.converters import to_native
from ansible.plugins.vars import BaseVarsPlugin
from ansible.utils.path import basedir
from ansible.inventory.group import InventoryObjectType
from ansible.utils.vars import combine_vars

from yamlvault.django import load_settings_from_config
from yamlvault.utils import flatten_list

CANONICAL_PATHS = {}  # type: dict[str, str]
FOUND = {}  # type: dict[str, list[str]]
NAK = set()  # type: set[str]
PATH_CACHE = {}  # type: dict[tuple[str, str], str]


class VarsModule(BaseVarsPlugin):
    allow_extras = True
    REQUIRES_ENABLED = False
    is_stateless = True

    def load_found_files(self, loader, data, found_files):
        for found in found_files:
            new_data = loader.load_from_file(found, cache="all", unsafe=True)
            if new_data:  # ignore empty files
                data = combine_vars(data, new_data)
        return data

    def parse_args(self):
        parser = argparse.ArgumentParser()
        parser.add_argument("playbook")  # positional argument
        parser.add_argument("-i", "--inventory")  # option that takes a value
        parser.add_argument("-l", "--limit")
        parser.add_argument("--vault-password-file", action="append", nargs="+")
        args, unknown = parser.parse_known_args()
        return args

    def get_vars(self, loader, path, entities, cache=True):
        """parses the inventory file"""

        if not isinstance(entities, list):
            entities = [entities]

        # realpath is expensive
        try:
            realpath_basedir = CANONICAL_PATHS[path]
        except KeyError:
            CANONICAL_PATHS[path] = realpath_basedir = os.path.realpath(basedir(path))

        args = self.parse_args()
        config_file_path, origin = self.get_option_and_origin("config_file_path")
        full_config_file_path = os.path.join(realpath_basedir, config_file_path)

        data = {}

        for entity in entities:
            # print(entity)

            try:
                entity_name = entity.name
            except AttributeError:
                raise AnsibleParserError(
                    "Supplied entity must be Host or Group, got %s instead"
                    % (type(entity))
                )

            try:
                first_char = entity_name[0]
            except (TypeError, IndexError, KeyError):
                raise AnsibleParserError(
                    "Supplied entity must be Host or Group, got %s instead"
                    % (type(entity))
                )

        vault_password_files = flatten_list(args.vault_password_file)
        data = load_settings_from_config(
            full_config_file_path, args.limit, password_files=vault_password_files
        )

        return data



import os
import io
import logging
import pprint

import ruamel.yaml as ruml
from ruamel.yaml import CommentedMap, CommentedSeq
from ruamel.yaml.scalarfloat import ScalarFloat
from ruamel.yaml.nodes import ScalarNode

from .environ_helper import generate_dynamic_environ
from .exceptions import (
    NoDefaultMapTagDefinedException,
    TooManyDefaultMapTagsDefinedException,
    EnvironmentNotFound,
    NoEnvironmentsDefinedException,
    EnvironmentIsAlreadyEncrypted,
    EnvironmentIsAlreadyDecrypted,
    EnvironmentHasNoSecretTagsException,
    VersionTagNotSpecified,
    UnsupportedVersionSpecified,
    UnsupportedEncryptionMethodSpecified,
)

from .tags import (
    DefaultSecretConfigMap,
    EnvSecretConfigMap,
    EncryptedString,
    SecretString,
    RequiredString,
    OtherScalar,
    StyledScalar, scalar_representer, scalar_constructor
)


from .constants import CONTAINS_ENCRYPTED_TAGS, CONTAINS_UNENCRYPTED_TAGS, VALID_ENCRYPTION_METHODS, ANSIBLE_VAULT_ENCRYPTION_METHOD
from .utils import deep_update, encrypt_value, decrypt_value



logger = logging.getLogger("eyaml")

class SecretYAML(ruml.YAML):
    def __init__(self, *args, filepath=None, **kwargs):
        super().__init__(*args, **kwargs)
        # self.indent(mapping=4, sequence=4, offset=2)
        self.width = 100000
        self.filepath = filepath
        self.data = None
        self.status = None
        self.encryption_method = ANSIBLE_VAULT_ENCRYPTION_METHOD # defaults to ansible vault

        # ruamel yaml settings
        self.register_class(DefaultSecretConfigMap)
        self.register_class(EnvSecretConfigMap)
        self.register_class(EncryptedString)
        self.register_class(SecretString)
        self.register_class(RequiredString)
        self.register_class(OtherScalar)
        self.constructor.add_constructor("tag:yaml.org,2002:str", scalar_constructor)
        self.representer.add_representer(StyledScalar, scalar_representer)

        self.default_flow_style = True
        self.preserve_quotes = True
        self.explicit_start = False
        self.anchor = None  # This tells ruamel.yaml to not use anchors

        if self.filepath:
            self.data = self.load_file(self.filepath)
        if self.data:
            self.validate()
            self.data = self._set_parent_walk(self.data, parent_name='__root__')

    @classmethod
    def get_tags_of_type(cls, node, of_type, depth=0):
        tags = {}
        if isinstance(node, dict):
            for key, value in node.items():
                if isinstance(key, of_type):
                    tags[key] = value
                tags.update(cls.get_tags_of_type(value, of_type, depth=depth + 1))

        elif isinstance(node, list):
            for item in node:
                tags.update(cls.get_tags_of_type(item, of_type, depth=depth + 1))

        return tags

    @classmethod
    def contains_tag_of_type(cls, node, of_type, depth=0):
        found = False
        if isinstance(node, dict):
            for key, value in node.items():
                if isinstance(key, of_type):
                    found = True
                    break
                found = cls.contains_tag_of_type(value, of_type, depth=depth + 1)
                if found:
                    break
        elif isinstance(node, list):
            for item in node:
                found = cls.contains_tag_of_type(item, of_type, depth=depth + 1)
                if found:
                    break
        else:
            if isinstance(node, of_type):
                return True
        return found

    def validate(self):
        self.get_default()
        if len(self.envs) == 0:
            raise NoEnvironmentsDefinedException()
        self.version_check()
        self.encryption_spec_check()

    def version_check(self):
        version = self.to_dict().get("version", False)
        if not version:
            raise VersionTagNotSpecified()
        if str(version) != "1.0":
            raise UnsupportedVersionSpecified()

    def encryption_spec_check(self):
        encryption_method = self.to_dict().get("encryption_method", self.encryption_method)
        if str(encryption_method) not in VALID_ENCRYPTION_METHODS:
            raise UnsupportedEncryptionMethodSpecified(encryption_method)

        logger.debug(f'Using {encryption_method} encryption method')
        self.encryption_method = encryption_method

    def has_no_secret_tags(self, node):
        return not self.contains_tag_of_type(node, SecretString)

    def has_not_encrypted_tags(self, node):
        return not self.contains_tag_of_type(node, EncryptedString)

    def encrypt_env(self, env_name, password):
        node = self.get_env_by_name(env_name)
        if self.is_encrypted(node):
            raise EnvironmentIsAlreadyEncrypted()

        if self.has_no_secret_tags(node):
            raise EnvironmentHasNoSecretTagsException()

        self.encrypt_walk(node, password)
        logger.info(f"Encrypted environment {env_name}")

    def decrypt_env(self, env_name, password, raise_exception=True):
        node = self.get_env_by_name(env_name)
        if self.is_decrypted(node):
            if raise_exception:
                raise EnvironmentIsAlreadyDecrypted()

        self.decrypt_walk(node, password, raise_exception=raise_exception)
        logger.info(f"Encrypted environment {env_name}")

    def encrypt_default(self, password):
        node = self.get_default()
        self.encrypt_walk(node, password)

    def decrypt_default(self, password, raise_exception=True):
        node = self.get_default()
        self.decrypt_walk(node, password, raise_exception=raise_exception)

    def is_default_encrypted(self):
        node = self.get_default()
        return self.is_encrypted(node)

    def is_default_decrypted(self):
        node = self.get_default()
        return self.is_decrypted(node)

    def encrypt_walk(self, node, password):
        if isinstance(node, SecretString):
            encrypted_string = encrypt_value(node.value, password, node)
            node = EncryptedString(encrypted_string, style=node.style)
        elif isinstance(node, dict):
            for k, v in node.items():
                node[k] = self.encrypt_walk(v, password)
        elif isinstance(node, list):
            for idx, item in enumerate(node):
                node[idx] = self.encrypt_walk(item, password)
        return node

    def decrypt_walk(self, node, password, raise_exception=True):
        if isinstance(node, EncryptedString):
            decrypted_string = decrypt_value(
                node.value, password, node, raise_exception=raise_exception
            )
            node = SecretString(decrypted_string, style=node.style)
        elif isinstance(node, dict):
            for k, v in node.items():
                node[k] = self.decrypt_walk(
                    v, password, raise_exception=raise_exception
                )
        elif isinstance(node, list):
            for idx, item in enumerate(node):
                node[idx] = self.decrypt_walk(
                    item, password, raise_exception=raise_exception
                )
        return node

    def _set_parent_walk(self, node, parent=None, parent_name=None):
        if isinstance(node, EncryptedString):
            node.parent = parent
            node.parent_name = parent_name
            return node
        if isinstance(node, CommentedMap):
            for k, v in node.items():
                try:
                    node[k].parent = parent
                    node[k].parent_name = parent_name
                except:
                    pass
                node[k] = self._set_parent_walk(v, parent=node, parent_name=str(k))
            node.parent = parent
            node.parent_name = parent_name
            return node
        elif isinstance(node, CommentedSeq):
            for idx, item in enumerate(node):
                node[idx].parent = parent
                node[idx].parent_name = parent_name
                node[idx] = self._set_parent_walk(item, parent=node, parent_name=parent_name)
            node.parent = parent
            node.parent_name = parent_name
            return node
        elif isinstance(node, ScalarFloat):
            return node
        elif isinstance(node, ScalarNode):
            node.parent = parent
            node.parent_name = parent_name
            return node
        elif isinstance(node, StyledScalar):
            node.parent = parent
            node.parent_name = parent_name
            return node
        elif isinstance(node, SecretString):
            node.parent = parent
            node.parent_name = parent_name
            return node
        elif isinstance(node, EncryptedString):
            node.parent = parent
            node.parent_name = parent_name
            return node
        elif isinstance(node, int):
            return node

        node.parent = parent
        node.parent_name = parent_name
        return node


    def load_file(self, filepath):
        if not os.path.isfile(filepath):
            raise FileNotFoundError(filepath)

        with open(filepath, "r") as stream:
            data_str = stream.read()
            self.representer.ignore_aliases = lambda *data: True
            return self.load(data_str)

    def save_file(self, filepath=None):
        if filepath is None:
            filepath = self.filepath

        buf = io.StringIO()
        self.dump(self.data, buf)
        data = buf.getvalue()

        with open(filepath, "w") as stream:
            stream.write(data)

    def save(self, filepath=None):
        self.save_file(filepath=filepath)

    def load_yaml(self, stream):
        return self.load(stream)

    def get_default(self, node=None):
        if node is None:
            node = self.data
        base = self.get_tags_of_type(node, DefaultSecretConfigMap)
        base_count = len(base.keys())
        if base_count == 0:
            raise NoDefaultMapTagDefinedException()

        if base_count != 1:
            raise TooManyDefaultMapTagsDefinedException()

        default_env_key = list(base.keys())[0]
        return self.data[default_env_key]

    @property
    def envs(self):
        return self.get_tags_of_type(self.data, EnvSecretConfigMap)

    @property
    def env_names(self):
        return list(self.envs.keys())

    def get_env_by_name(self, name):
        mapping_by_str = {str(k): v for k, v in self.envs.items()}
        if name not in list(mapping_by_str.keys()):
            raise EnvironmentNotFound(msg=f"Environment of {name} not found")
        return mapping_by_str[name]

    def env(self, env_name):
        if env_name == "default":
            return self.get_default()
        return self.get_env_by_name(env_name)

    def get_env(self, name):
        return self.env(name)

    # def get_env_as_dict(self, name):
    #     return self.to_dict(self.get_env_by_name(name))

    def is_env_encrypted(self, name):
        return self.is_encrypted(self.get_env_by_name(name))

    def is_env_decrypted(self, name):
        return self.is_decrypted(self.get_env_by_name(name))

    def is_encrypted(self, node=None):
        if node is None:
            node = self.data
        if self.contains_tag_of_type(node, SecretString):
            logger.warning(CONTAINS_UNENCRYPTED_TAGS)
        return self.contains_tag_of_type(node, EncryptedString)

    def is_decrypted(self, node=None):
        if node is None:
            node = self.data
        if self.contains_tag_of_type(node, EncryptedString):
            pass
            # logger.warning(CONTAINS_ENCRYPTED_TAGS)

        return self.contains_tag_of_type(node, SecretString)

    def to_dict(self, node=None):
        if node is None:
            node = self.data
        if isinstance(node, EncryptedString):
            # TODO: raise warning on encrypted string being dict dumped
            return f"{node.value}"
        if isinstance(node, CommentedMap):
            return {f"{k}": self.to_dict(v) for k, v in node.items()}
        elif isinstance(node, CommentedSeq):
            return [self.to_dict(item) for item in node]
        elif isinstance(node, ScalarFloat):
            return node
        elif isinstance(node, ScalarNode):
            return f"{node}"
        elif isinstance(node, StyledScalar):
            return f"{node}"
        elif isinstance(node, SecretString):
            return f"{node.value}"
        elif isinstance(node, int):
            return node
        else:
            print(type(node))
            return node

    def get_default_as_dict(self):
        default_node = self.get_default()
        default_dict = self.to_dict(default_node)
        return default_dict

    def get_env_as_dict(self, env, use_default=True):
        default_node = self.get_default()
        env_node = self.get_env_by_name(env)
        env_node_dict = self.to_dict(env_node)
        if use_default:
            default_dict = self.to_dict(default_node)
            default_dict = deep_update(default_dict, env_node_dict)
            return default_dict
        return env_node_dict

    def patch_object_with_env(self, obj, env_name):
        env = self.get_env_as_dict(env_name)
        if isinstance(obj, dict):
            obj = deep_update(obj, env)
            return obj

        if str(type(obj)) == "<class 'django.conf.LazySettings'>":
            for k, v in env.items():
                setattr(obj, k, v)
                getattr(obj, k)
            return obj

    def patch_environs(self, env):
        obj = {}
        config = self.get_env_as_dict(env)
        return generate_dynamic_environ(config)

    def __repr__(self):
        pp = pprint.PrettyPrinter(depth=5, stream=None)
        return pp.pformat(self.data)
